/*
 *
 *                Copyright 1991,1992,1993 Premia Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Premia Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Premia Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Premia Corp.
 *
 *    Owner:
 *        Premia Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN    /* defines the LibMain() */


#define MAX_BUF_LEN    100

static LPVOID _wrl_hash_names        = NIL;    /* pointer to the hash table containing */
                                               /* all 'VRML' keyword names. */

static LPVOID _wrl_hash_names_1      = NIL;    /* pointer to optional hash table containing */
static int    _wrl_hash_size_1       = 47;     /* user defined keyword names and size of
                                               /* table. When created these key
                                               /* are set to Color Alternate5 */           
                        

static LPVOID _wrl_hash_names_2      = NIL;    /* pointer to optional hash table containing */
static int    _wrl_hash_size_2       = 47;     /* user defined keyword names and size of
                                               /* table. When created these key
                                               /* are set to Color Alternate6 */              

static LPVOID _wrl_hash_templates    = NIL;    /* pointer to the hash table containing */
                                               /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _wrl_hash_templates.
 */
typedef struct _LanguageTemplate {         
    LPSTR abbrev;
    LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR WRL_INDENT_STR = "\377DEF\377USE\377";

/*
 * _wrl_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _wrl_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _wrl_templates[] = {
//    Abbreviation                       Template
	// Geometry nodes
    {"Asc|AsciiText",              "AsciiText%0{\n%1&\n%2}" },   
    {"Cone",                       "Cone%0{\n%1&\n%2}" },   
    {"Cube",                       "Cube%0{\n%1&\n%2}" },   
   	{"Cyl|Cylinder",               "Cylinder%0{\n%1&\n%2}" },   
 	{"IFS|IndexedFaceSet",         "IndexedFaceSet%0{\n%1&\n%2}" },   
    {"ILS|IndexedLineSet",         "IndexedLineSet%0{\n%1&\n%2}" },   
 	{"PS|PointSet",                "PointSet%0{\n%1&\n%2}" },   
 	{"Sph|Sphere",                 "Sphere%0{\n%1radius &\n%2}" },   
	// Property nodes
    {"Coo|Coordinate3",            "Coordinate3%0{\n%1point &\n%2}" },   
    {"Font|FontStyle",             "FontStyle%0{\n%1&\n%2}" },   
    {"Inf|Info",                   "Info%0{\n%1string \"&\"\n%2}" },   
    {"LOD",                        "LOD%0{\n%1&\n%2}" },   
    {"Mat|Mate|Material",          "Material%0{\n%1&\n%2}" },   
    {"MatB|MaterialBinding",       "MaterialBinding%0{\n%1&\n%2}" },   
    {"Nor|Normal",                 "Normal%0{\n%1&\n%2}" },   
    {"NorB|NormalBinding",         "NormalBinding%0{\n%1&\n%2}" },   
    {"Tex|Texture2",               "Texture2%0{\n%1&\n%2}" },   
    {"Textr|Texture2Transform",    "Texture2Transform%0{\n%1&\n%2}" },   
  	{"TeC|TextureCoordinate2",     "TextureCoordinate2%0{\n%1point &\n%2}" },   
	{"ShapeHints",                 "ShapeHints%0{\n%1&\n%2}" },   
    {"MaTran|MatrixTransform",     "MatrixTransform%0{\n%1matrix &\n%2}" },   
  	{"Rot|Rotation",               "Rotation%0{\n%1rotation &\n%2}" },   
    {"Scale",                      "Scale%0{\n%1scaleFactor &\n%2}" },   
    {"Transform",                  "Transform%0{\n%1&\n%2}" },   
    {"Translation",                "Translation%0{\n%1translation &\n%2}" },   
    {"Orth|OrthographicCamera",    "OrthographicCamera%0{\n%1&\n%2}" },   
 	{"PerspectiveCamera",          "PerspectiveCamera%0{\n%1&\n%2}" },   
    {"DL|DirectionalLight",        "DirectionalLight%0{\n%1&\n%2}" },   
 	{"PL|PointLight",              "PointLight%0{\n%1&\n%2}" },   
    {"SL|SpotLight",               "SpotLight%0{\n%1&\n%2}" },   
	// Group nodes
	{"Group",                      "Group%0{\n%1&\n%2}" },   
 	{"Sep|Separator",              "Separator%0{\n%1&\n%2}" },   
 	{"Switch",                     "Switch%0{\n%1whichChild &\n%2}" },   
 	{"TranSep|TransformSeparator", "TransformSeparator%0{\n%1&\n%2}" },   
 	{"WWWA|WWWAnchor",             "WWWAnchor%0{\n%1&\n%2}" },   
	// Outcast nodes
	{"WWWI|WWWInline",             "WWWInline%0{\n%1&\n%2}" },   

	{"{",                          "%0{\n%1&\n%2}" },   
	{"[",                          "[ & ]" },
    { NIL, NIL    },
};

/* Define the list of 'VRML' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _wrl_names[] = {
	// Keywords
   "SFBitMask", "SFBool",   "SFColor",    "SFEnum",   "SFFloat", "SFImage",
   "SFLong",    "SFMatrix", "SFRotation", "SFString", "SFVec2f", "SFVec3f",
   "MFColor",   "MFLong",   "MFString",   "MFVec2f",  "MFVec3f", "fields",
   "DEF",       "USE",
   // Constants
   "OVERALL",            "TRUE",               "FALSE",            "LEFT",             
   "CENTER",             "RIGHT",              "SIDES",            "BOTTOM",           
   "ALL",                "PER_PART",           "PER_PART_INDEXED", "PER_FACE",         
   "PER_FACE_INDEXED",   "TOP",                "SERIF",            "SANS",             
   "TYPEWRITER",         "BOLD",               "ITALIC",           "PER_VERTEX",       
   "PER_VERTEX_INDEXED", "DEFAULT",            "UNKNOWN_ORDERING", "CLOCKWISE",        
   "COUNTERCLOCKWISE",   "UNKNOWN_SHAPE_TYPE", "SOLID",            "UNKNOWN_FACE_TYPE",
   "CONVEX",             "REPEAT",             "CLAMP",            "NONE",
   "POINT",             

    NIL
};

/* Define the list of 'VRML' native nodes to appear in
 * the second keyword color if language coloring is enabled.
 */
LPSTR _wrl_nodes[] = {
	// Geometry nodes
    "AsciiText",      "Cone",     "Cube",   "Cylinder", "IndexedFaceSet",
    "IndexedLineSet", "PointSet", "Sphere",
	// Property nodes
    "Coordinate3",      "FontStyle",          "Info",               "LOD",           
    "Material",         "MaterialBinding",    "Normal",             "NormalBinding", 
    "Texture2",         "Texture2Transform",  "TextureCoordinate2", "ShapeHints",   
    "MatrixTransform",  "Rotation",           "Scale",              "Transform",    
    "Translation",      "OrthographicCamera", "PerspectiveCamera", 
    "DirectionalLight", "PointLight",         "SpotLight",         
	// Group nodes
	"Group", "Separator", "Switch", "TransformSeparator", "WWWAnchor",
	// Outcast nodes
	"WWWInline",
    NIL
};


static long _wrl_in_comment( long offset );

void DLL _wrl_init( void );
BOOL DLL _wrl_expand( void );
void DLL _wrl_language_colors_update( long firstline, long lastline );
void DLL _wrl_language_colors( long numlines );
void DLL _wrl_assign_template( LPSTR abbrev, LPSTR template );
void DLL _wrl_add_keyword( LPSTR keyword, int table );
BOOL DLL _wrl_indent( void );
BOOL DLL _wrl_insert_brace( void );
void DLL _wrl_hash_size( int ht, int size );
void DLL _wrl_routines( void);


void DLL 
_init()
{
    _wrl_init();
}

/*
 ** _wrl_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_wrl_init( void )
{
    static BOOL firstTime = TRUE;
    int    i;

    if (firstTime)
    {
        /* _wrl language handling functions get called via LibFunctionExec,
         * so they must be exported.
         */
        LibExport( "BOOL _wrl_expand");
        LibExport( "_wrl_assign_template LPSTR LPSTR");
        LibExport( "BOOL _wrl_indent");
		LibExport( "BOOL _wrl_insert_brace");
        LibExport( "_wrl_language_colors_update long long");
        LibExport( "_wrl_language_colors long");
        LibExport( "_wrl_add_keyword LPSTR int");
        LibExport( "_wrl_hash_size int int");
		LibExport( "_wrl_routines" );

        firstTime = FALSE;

        /* Create a hash table to hold all of the 'VRML' keyword names.
         */
        _wrl_hash_names = HashCreateTable( 29 /* a prime # */, sizeof( LPSTR ), 0 );
        i = 0;

        while ( _wrl_names[ i ] != NIL )
        {
            LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _wrl_hash_names, 
                    _wrl_names[i], strlen( _wrl_names[i] ) );
            if (lpf)
            {
                *lpf = StrNew( _wrl_names[i] );
                i++;
            }
        }

        /* Create a hash table to hold all of the 'VRML' native nodes.
         */
        _wrl_hash_names_1 = HashCreateTable( 29, sizeof( LPSTR ), 0 );
        i = 0;

        while ( _wrl_nodes[ i ] != NIL )
        {
            LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _wrl_hash_names_1, 
                    _wrl_nodes[i], strlen( _wrl_nodes[i] ) );
            if (lpf)
            {
                *lpf = StrNew( _wrl_nodes[i] );
                i++;
            }
        }

        /* Create a hash table to hold all of the 'VRML' template names.
         */
        _wrl_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _wrl_templates[ i ].abbrev != NIL )
        {
            _wrl_assign_template( _wrl_templates[i].abbrev, _wrl_templates[i].template );
            i++;
        }
    }
	CompilerNewExt( ".wrl" );

}


void DLL
_wrl_hash_size( int ht, int size )
{
	if (size > 0)
	{
		if (ht == 1)
			_wrl_hash_size_1 = size;
		else if (ht == 2)
			_wrl_hash_size_2 = size;
	}
}


/*
 ** _wrl_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'VRML' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_wrl_expand( void )
{
    LPVOID htEntry;
    LPSTR  _wrl_string = NULL;
    LPSTR  str;
    int    i = 0;

    MarkSavePos();

    /* Check to see if we are currently positioned inside a 'VRML' comment
     */
    if (_wrl_in_comment( -1L ) != -1)
        goto _wrl_expand_incomplete;

    str = BufReadStr( 65535 );
    while (str[i] && !isspace(str[i]))
        i++;
    str[i] = '\0';

    _wrl_string = StrTrim( str, " \t" );
    StrFree( str );

    if (!_wrl_string)
        goto _wrl_expand_incomplete;

    /* Search for a keyword match in the hash table.
     */
    htEntry = HashFindEntry( _wrl_hash_templates, _wrl_string, strlen( _wrl_string ) );
    StrFree( _wrl_string );

    if (htEntry)
    {
        /* A template exists for the current keyword, delete the keyword
         * and insert the template.
         */
        MarkDropPos( 1 );
        str = *(LPSTR*)htEntry;
        BufDelToEOL();

        _ext_expand_template( str );

    } 
    else
        goto _wrl_expand_incomplete;
    return TRUE;

_wrl_expand_incomplete:
    MarkRestorePos();

    return FALSE;
}

/*
 ** _wrl_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
void DLL
_wrl_language_colors_update( long firstline, long lastline )
{
    long startoff;

    if (firstline && lastline)
    {
        /* Begin coloring at the previous comment if
         * currently located in the middle of one.
         */
        _PosInit( 0 );
        if (firstline == 1)
            _PosPrevLine( 0L );
        else
            _PosNextLine( firstline - 1);
        startoff = _PosQOffset();

        /* Clear all attributes for the line range and
         * the reset them within _wrl_language_colors.
         */
        AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
        _PosInit( startoff );
        _wrl_language_colors( lastline - firstline + 1 );
    }
}
                    
void DLL
_wrl_language_colors( long numlines )
{
    long  startoff;      
    long  startline;     
    long  endoff;        
    long  inc;           
    WORD  ch;            
    LPSTR buf;           
    int   i               = 0;              
    int   comment_color   = ColorComments(   -1 );
    int   keyword_color   = ColorKeywords(   -1 );
    int   quote_color     = ColorAlternate1( -1 );
    int   number_color    = ColorAlternate2( -1 );
	int   brace_color     = ColorAlternate3( -1 );
    int   keyword_color_1 = ColorAlternate5( -1 );
    int   keyword_color_2 = ColorAlternate6( -1 );

    buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

    /* Start at the top of the buffer or current position 
     */
    if (numlines == -1)
    {
        _PosInit( 0 );  
        numlines = 0x7FFFFFFF;
    }
    else
    {
        startoff  = _PosQOffset();
        startline = _PosQLine();
        endoff    = _wrl_in_comment( startoff );
        if (endoff == -1)
            _PosInit( startoff );
        else
        {
            _PosInit( endoff );
            numlines += (startline - _PosQLine());
        }
    }
    ch = _PosCurrentChar();
    startoff = _PosQOffset();


    while ((ch != EOF_CHAR) && (numlines > 0))
    {
        if (ch == '\n')
            numlines--;

        if (ch == '"' || ch == '\'')
        {
            UINT quote = ch;

            startoff = _PosQOffset();

            do 
            {
                ch = _PosNextChar();
                if (ch == '\\')
                {
                    ch = _PosNextChar();
                    if ((ch == quote) || (ch == '\\'))
                        ch = _PosNextChar();
                }
                if ((ch == EOF_CHAR) || (ch == '\n'))
                    break;
            } while (ch != quote);

            if (ch == quote)
                ch = _PosNextChar();

            /* color the string */
            if (quote_color)
            {
                endoff = _PosQOffset();
                _PosInit( startoff );
                _PosSetColor( quote_color, endoff - startoff );
                _PosInit( endoff );
            }
            continue;
        }
        else if (number_color && isdigit( ch ))
        {
            int j = 0;
            UINT lastch = _PosPrevChar();

            if (lastch != EOF_CHAR)
                _PosNextChar();

            startoff = _PosQOffset();
            do
            {
                j++;
                ch = _PosNextChar();
            } while (isxdigit( ch ) || (ch == '.') || (ch == 'x' && j == 1));

            if ((  lastch != ' ' 
                && lastch != '\t' 
                && lastch != '\r' 
                && lastch != '\n' 
                && lastch != EOF_CHAR
                && !ispunct( lastch ))
                || lastch == '_'
                || (ch == '_'))
                continue;

            if (ch == 'l' || ch == 'L')
                ch = _PosNextChar();

            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( number_color, endoff - startoff );
            _PosInit( endoff );

        }
        else if (ISALNUM( (BYTE)ch ) || (ch == '_') || (ch == '.'))
        {
            WORD lastch = _PosPrevChar();

            if (lastch != EOF_CHAR)
                _PosNextChar();

            i = 0;
            startoff = _PosQOffset();

            do
            {
                if (i < MAX_BUF_LEN)
                    buf[i++] = (char)ch;
                ch = _PosNextChar();
            } while (ISALNUM( (BYTE)ch ) || (ch == '_') || (ch == '.'));

            if (STRCHR( " \t\r\n[(:", lastch ) == NULL)
                i = 0;

            continue;
        }

		else if (brace_color && STRCHR("{}", ch) != 0)    
		{                                        
			long coff;                            

			coff = _PosQOffset();            
			_PosInit(coff);                  
			_PosSetColor(brace_color, 1);    
			_PosInit(coff);                  
		}                                   

        else if (i)
        {
            if (ISALNUM( (BYTE)buf[0] ))
            {
                if (HashFindEntry( _wrl_hash_names, buf, i ))
                {
                    endoff = _PosQOffset();
                    _PosInit( startoff );
                    _PosSetColor( keyword_color, endoff - startoff );
                    _PosInit( endoff );
                }
                else if (HashFindEntry( _wrl_hash_names_1, buf, i ))
                {
                    endoff = _PosQOffset();
                    _PosInit( startoff );
                    _PosSetColor( keyword_color_1, endoff - startoff );
                    _PosInit( endoff );
                }
                else if (HashFindEntry( _wrl_hash_names_2, buf, i ))
                {
                    endoff = _PosQOffset();
                    _PosInit( startoff );
                    _PosSetColor( keyword_color_2, endoff - startoff );
                    _PosInit( endoff );
                }
            }
            i = 0;
        } 

        if (ch == '#')
        {
            startoff = _PosQOffset();

            inc = _PosNextLine( 1 );
            numlines--;
            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( comment_color, endoff - startoff - inc );
            _PosInit( endoff );
            ch = _PosCurrentChar();
            continue;
        }
        ch = _PosNextChar();
    }
    MemFree( buf );
}


/*
 ** _wrl_add_keyword()
 *
 *  PARAMETERS:
 *    keyword - the keyword to add to the user-defined keyword table
 *    table   - the 1-based index of the table to which to add the keyword
 *
 *  DESCRIPTION:
 *    Provide a means of adding user-defined keywords for special
 *    coloring
 */
void DLL
_wrl_add_keyword( LPSTR keyword, int table )
{
    if (keyword && *keyword && table >= 0)
    {
        LPSTR XFAR *lpf;
        LPVOID  ht;

        switch( table )
        {
        case 0:
            ht = _wrl_hash_names;
            break;
        case 1:
            if (_wrl_hash_names_1 == NULL)
                _wrl_hash_names_1 = HashCreateTable( _wrl_hash_size_1, sizeof( LPSTR ), 0 );
            ht = _wrl_hash_names_1;
            break;
        case 2:
            if (_wrl_hash_names_2 == NULL)
                _wrl_hash_names_2 = HashCreateTable( _wrl_hash_size_2, sizeof( LPSTR ), 0 );
            ht = _wrl_hash_names_2;
            break;
        }

        lpf = (LPSTR XFAR *)HashGetEntry( ht, keyword, strlen( keyword ) );
        if (lpf)
        {
            *lpf = StrNew( keyword );
        }
    }
}


/*
 ** _wrl_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'VRML' template strings
 *    without changing the code.
 */
void DLL
_wrl_assign_template( LPSTR abbrev, LPSTR template )
{
    _ext_assign_template( _wrl_hash_templates, abbrev, template );
}


/*
 ** _wrl_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'VRML' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in VRML_INDENT_STR - perform smart
 *       indentation.  
 */
BOOL DLL
_wrl_indent( void )
{
    long   curline;
    long   col;
    LPSTR  _wrl_string;
    LPSTR  str;
    long   matchpos;
 	WORD   endch;                                                            

    MarkSavePos();
    curline = BufQCurrentLine();

    /* Check to see if we are currently positioned inside a 'VRML' comment
     */
    if (_wrl_in_comment( -1L ) != -1)
        goto _wrl_indent_incomplete;

	/* Check to see if there is a trailing '{' on the current line,          
	 * and if so, indent on the next line.                                   
	 */                                                                      

	do {                                                                     
		if (MovPrevChar( 1L ) == FALSE)                                      
			break;                                                           
		endch = BufReadChar();                                               
	} while (endch == ' ' || endch == '\t');                                 
	MarkRestorePos();                                                        
	if (endch == '{')                                                        
		goto _wrl_indent_next_line;

 	/* Check to see if a '{' is by itself on the next line. If so, skip      
 	 * it and indent on the line following it.                                
  		 */                                                              
    MarkSavePos();
    curline = BufQCurrentLine();
 	if (SrchFind( "^[ \t]*{\\c[ \t]*$", SEARCH_FORWARD | SEARCH_REGEX, NIL ) 
 			&& (BufQCurrentLine() == curline + 1) )                          
 		MarkDropPos( 1 );                                                    
 	else
 	    {                                                                 
 		MarkRestorePos();                                                    
		goto _next_test;
		}
                                                                           
_wrl_indent_next_line:                                                    
 	_c_indent_nl();                                                          
 	return TRUE;        
                      
_next_test:
    /* Not in a comment so indent according to the current 'VRML' construct.
     */
    MarkSavePos();
    if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
            || (BufQCurrentLine() != curline) )
        goto _wrl_indent_incomplete;

    /* Read the keyword at the beginning of the current line.
     */
    col = BufQCurrentCol();
    SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
    str = BufReadStr( col - BufQCurrentCol() );
    _wrl_string = StrTrim( str, " \t" );
    StrFree( str );

    /* Set prefix and suffix to \377 to perform a quick lookup of the name
     */
    str = StrNew( "\377" );
    _wrl_string = StrAppend( _wrl_string, str );
    str = StrAppend( str, _wrl_string );
    StrFree( _wrl_string );
    
    matchpos = StrMatch( str, WRL_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
    StrFree( str );

    if (!matchpos)
        /* Not a 'VRML' keyword, do normal indentation. */
        goto _wrl_indent_incomplete;
        
    MarkRestorePos();

    _ext_insert_indented_EOL();
    BufInsertChar( '\t' );
    return TRUE;

_wrl_indent_incomplete:
        MarkRestorePos();

    return FALSE;

/////////////////////////////////////////////////////////////////////////////
// 	long   curline;                                                          
// 	WORD   endch;                                                            
//                                                                           
// 	MarkSavePos();                                                           
//                                                                           
// 	curline = BufQCurrentLine();                                             
//                                                                           
// 	/* Check to see if we are currently positioned inside a 'VRML' comment   
// 	 */                                                                      
// 	if (_wrl_in_comment( -1L ) != -1)                                        
// 		goto _wrl_indent_incomplete;                                         
//                                                                           
// 	/* Check to see if there is a trailing '{' on the current line,          
// 	 * and if so, indent on the next line.                                   
// 	 */                                                                      
// 	do {                                                                     
// 		if (MovPrevChar( 1L ) == FALSE)                                      
// 			break;                                                           
// 		endch = BufReadChar();                                               
// 	} while (endch == ' ' || endch == '\t');                                 
// 	MarkRestorePos();                                                        
// 	if (endch == '{')                                                        
// 		goto _wrl_indent_next_line;                                          
// 	MarkSavePos();                                                           
//                                                                           
// 	/* Not in a comment so indent according to the current 'VRML' construct. 
// 	 */                                                                      
//                                                                           
// 	/* Check to see if a '{' is by itself on the next line. If so, skip      
// 	 * it an indent on the line following it.                                
//  		 */                                                              
// 	if (SrchFind( "^[ \t]*{\\c[ \t]*$", SEARCH_FORWARD | SEARCH_REGEX, NIL ) 
// 			&& (BufQCurrentLine() == curline + 1) )                          
// 		MarkDropPos( 1 );                                                    
// 	else                                                                     
// 		MarkRestorePos();                                                    
//                                                                           
// _wrl_indent_next_line:                                                    
// 	_c_indent_nl();                                                          
// 	return TRUE;                                                             
//                                                                           
// _wrl_indent_incomplete:                                                   
// 	MarkRestorePos();                                                        
//                                                                           
// 	return FALSE;                                                            
/////////////////////////////////////////////////////////////////////////////
}

/*
 ** _wrl_routines()
 *
 *  DESCRIPTION:
 *    Locates function definitions for selective display mode.
 *    This function gets called by the Codewright API call Routines()
 *    if the current file is a .VRML file.
 *    The code here is taken from _c_routines; modify it as needed.
 */
void DLL
_wrl_routines( void )
{
	int   matches = 0;
	LPSTR str     = "^[a-zA-Z0-9_].*\\((.|\n)*\\)";
	LPSTR stre    = "^[a-zA-Z0-9_].*\\((.|\n)*\\)[ \t\n]*\\c[^ \t\n]";
	long  curline;
	UINT  ch;
	long  lines;

	SysWaitStart();

	MarkSavePos();

	BufSetAscii();
	MovTopBuf();

	while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD, 0 ))
	{
		curline = BufQCurrentLine();

		SrchFind( stre, SEARCH_REGEX | SEARCH_FORWARD, 0 );

		ch = BufReadChar();
		if (ch != ';')
		{
			if (!matches)
				AttrSetVisible( 0, 0, 0 );
			matches++;
			lines = BufQCurrentLine() - curline
						- !!(ch == '{');
			AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
			MovHome();
		}

		if (!MovDown( 1 ))
			break; 
	}

	if (matches)
	{
		BufSetCompact();
		MarkDropPos( 1 );
		MovTopBuf();
	}
	else
		MarkRestorePos();

	SysWaitEnd();

	MsgMessage( matches ? "" : "No routines found." );
}

PUBLIC BOOL
_wrl_in_string( long offset )
{
	UINT  ch;
	long  quotes     = 0;

	/* Check to see if we are currently positioned inside a 'C' string
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	while (TRUE)
	{
		do
		{
			if (ch == '"')
			{
				//ch = _PosPrevChar();
				quotes++;
			}
			ch = _PosPrevChar();
		} while ((ch != EOF_CHAR) && (ch != '\n'));

		if (ch == '\n')
			ch = _PosPrevChar();
	 	if (ch == '\r')
			ch = _PosPrevChar();
		if (ch != '\\')
			break;
		else
			ch = _PosPrevChar();
	}
	return ((int)quotes & 1);
}

static LPSTR 
ReadLine(void)
{
	LPSTR pLine, p;
	long  colFirst = BufQCurrentCol();

	pLine = StrNNew("", (int)colFirst);
	_fmemset(pLine, ' ', (int)colFirst);
	pLine[colFirst] = 0;
	pLine = StrAppend(pLine, p = BufReadStr(32767));
	StrFree(p);
	return(pLine);
}

static int 
firstChar(LPSTR str,long pos)
{
	/*
	**	Desc: Returns the first non-white character of the string
	**			before position pos or '\0'if none
	**	 Pre: Input string
	**	Post: first char
	** History:
	**	01-Oct-1990 JVV: Created
	*/

	LPSTR pTmp;
	LPSTR endpoint;


	pTmp = str;
	endpoint = str + pos;
	while ( *pTmp && pTmp < endpoint && (*pTmp == ' ' || *pTmp == '\t'))
		pTmp++;

	if (!(*pTmp) || pTmp == endpoint)
		return 0;

	return(*pTmp);
}

/**************************************************************************/

static int 
lastChar(LPSTR str,long pos)
{
	/*
	**	Desc: Finds the last non whitespace character before the position pos
	**	 Pre: Input string and position
	**	Post: last nonblank char
	** History:
	**	01-Oct-1990 JVV: Created
	*/

	LPSTR pTmp;

	pTmp = str + min(((int)pos),(int)(STRLEN(str) - 1));

	while (pTmp >= str && (*pTmp == ' ' || *pTmp == '\t'))
		pTmp--;

	if (pTmp < str)
		return 0;

	return *pTmp;
}

BOOL DLL
_wrl_insert_brace( void )
{
	LPVOID htEntry;
	int fEmptyLine;
	LPSTR buf;
	long xTarg, yTarg;

	xTarg = BufQCurrentCol();
	yTarg = BufQCurrentLine();

	if ((_wrl_in_comment(-1) == -1) &&
		(!_wrl_in_string(-1)))
	{

		if (xTarg == 1)
		{ /* function delimiter */
			_ext_expand_template( "{\n%1&\n%home}" );
		}
		else
		{
			MovAbs(0, 1);
			buf = ReadLine();
			fEmptyLine = (firstChar(buf, STRLEN(buf)) == 0);
			StrFree(buf);
			MovAbs(yTarg, xTarg);
		
			if (fEmptyLine)
				if (MovUp( 1 ))
					MovEOL();
		
			_wrl_init();
			htEntry = HashFindEntry( _wrl_hash_templates, "{", 1 );

			if (htEntry)
				_ext_expand_template( *(LPSTR*)htEntry );
			else
				BufInsertChar( 0x7b );
		}
	}
	else
		BufInsertChar( 0x7b );

	return TRUE;
}

/*--  Private functions --*/


/*
 ** _wrl_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'VRML' comment by search backward for a # 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_wrl_in_comment( long offset )
{
    WORD  ch;

    /* Check to see if we are currently positioned inside a 'VRML' comment
     */
    _PosInit( offset );
    ch = _PosCurrentChar();

    do
    {
        if (ch == '\n')
            return -1;
        else if (ch == '#')
            return _PosQOffset();  /* currrent line is in a comment */
        ch = _PosPrevChar();
    } while (ch != EOF_CHAR);

    return -1L;
}

